import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.PrintStream;
import java.util.ArrayList;

import org.semanticweb.owlapi.model.OWLAxiom;



import edu.ntua.ivml.qa.algorithm.ExecutionResult;
import edu.ntua.ivml.qa.algorithm.Executor;
import edu.ntua.ivml.qa.algorithm.rapid.Rapid;
import edu.ntua.ivml.qa.dllite.DLLiteRTheory;
import edu.ntua.ivml.qa.dllite.OWL2ClauseSet;
import edu.ntua.ivml.qa.dllite.OntologyReference;
import edu.ntua.ivml.qa.lp.Clause;
import edu.ntua.ivml.qa.lp.ClauseParser;
import edu.ntua.ivml.qa.utils.Utils;


public class Terminal {

	private static DLLiteRTheory lp;
	private static Clause query; 
	
	public static void main(String[] args) throws Exception {

		if (args.length < 4) {
			System.out.println("Syntax : Terminal mode[N/F/D] ontologyURL queryFile outputFile");
			System.out.println("                       N will run Rapid");
			System.out.println("                       F will run Rapid-f (optimized subsumption checking)");
			System.out.println("                       D will run Rapid-d (generates a datalog program)");
			System.out.println();
			return;
		}
		
		String mode = args[0];			// N will run Rapid      
										// F will run Rapid-f (optimized subsumption checking)
										// D will run Rapid-d (the output will be a datalog program)
		String ontologyFile = args[1];
		String queryFile = args[2];
		String outputFile = args[3];

		PrintStream out = new PrintStream(new FileOutputStream(new File(outputFile)));
		
		OntologyReference ontRef = new OntologyReference(ontologyFile);
		
		out.println("Ontology : " + ontologyFile + " (" + ontRef.getOntology().getAxiomCount() + " axioms)");
		out.println("Query    : " + queryFile);
		out.println();
		
		OWL2ClauseSet dl2lp = new OWL2ClauseSet(ontRef);
		
		lp = new DLLiteRTheory(dl2lp.rewrite());
		
		ArrayList<OWLAxiom> ignoredAxioms = dl2lp.getIgnoredAxioms();
		if (ignoredAxioms.size() > 0) {
			Utils.printObjectArrayList(out, "IGNORED AXIOMS", ignoredAxioms);
		}
		
		query = (new ClauseParser()).parseClause((new BufferedReader(new FileReader(queryFile))).readLine()); 

		Utils.printObjectArrayList(out, " LOGIC PROGRAM ", lp.getClauses());
		out.println("QUERY : " + query);
		out.println();
		
		Rapid engine = null;
		
		if (mode.equals("N")) {
			engine = Rapid.createRapid(1);
		} else if (mode.equals("F")) {
			engine = Rapid.createRapid(2);	
		} else if (mode.equals("D")) {
			engine = Rapid.createRapidDatalog();			
		}
		
		out.println("-------------------------------------------------------------");
		out.println("REASONER : " + engine.getEngineName() );
		out.println("-------------------------------------------------------------");
		
		engine.setTheory(lp);
			
		Executor exec = new Executor(engine);
		
		ExecutionResult res = exec.threadExecute(query, out);
			
		ArrayList<Clause> rewritings = res.getRewritings();

		out.println("Reasoning time         : " + exec.reasoningTime + " ms.");
		out.println("Redundacy removal time : " + exec.redundantRemovalTime + " ms.");
		out.println("Total time             : " + (exec.reasoningTime + exec.redundantRemovalTime) + " ms.");
		out.println();

		out.println("Rewriting size         : " + exec.allRewritingsSize);
		out.println("Core rewriting size    : " + exec.coreRewritingsSize);
		out.println();

		Utils.printObjectArrayList(out, " CORE REWRITINGS ", rewritings);
	}
	
}


