package edu.ntua.ivml.qa.algorithm.rapid;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import edu.ntua.ivml.qa.algorithm.Engine;
import edu.ntua.ivml.qa.dllite.DLLiteRTheory;
import edu.ntua.ivml.qa.lp.Atom;
import edu.ntua.ivml.qa.lp.Clause;
import edu.ntua.ivml.qa.lp.ClauseContainer;
import edu.ntua.ivml.qa.lp.Constant;
import edu.ntua.ivml.qa.lp.Function;
import edu.ntua.ivml.qa.lp.Predicate;
import edu.ntua.ivml.qa.lp.Substitution;
import edu.ntua.ivml.qa.lp.Term;
import edu.ntua.ivml.qa.lp.Tuple;
import edu.ntua.ivml.qa.lp.Variable;
import edu.ntua.ivml.qa.utils.SetUtils;




public class Rapid extends Engine {
	private AtomUnfolder au;
	
	private int mode;
	private int mark;
	
	private boolean datalog;
	
	public static Rapid createRapidDatalog() {
		return new Rapid(1, true);
	}

	public static Rapid createRapid(int mode) {
		return new Rapid(mode, false);
	}

	private Rapid(int mode, boolean datalog) {
		super("Rapid" + (datalog ? "-d" : (mode == 1 ? "" : "-f")));
		this.datalog = datalog;
		this.mode = mode;
	}
	
	public void setTheory(DLLiteRTheory lp) {
		super.setTheory(lp);
		
		au = new AtomUnfolder(lp);
	}

	public AtomUnfolder getAtomUnfolder() {
		return au;
	}
	
	public void execute() {
		
		mark = 0;

		AtomUnfolder.mode = mode;

		UnfoldedClause uQuery = (UnfoldedClause)new UnfoldedClause(query).removeRedundantAtoms();

		ArrayList<MarkedUnfoldedClause> queryList = new ArrayList<MarkedUnfoldedClause>();
		queryList.add(new MarkedUnfoldedClause(uQuery, mark));

		int j = 0;
		while (j < queryList.size()) {
			resolve(queryList, queryList.get(j++).getClause());
		}
		
		if (datalog) {
			expand(queryList);
		} else {
			unfold(queryList);
		}
	}
	
	private void resolve(ArrayList<MarkedUnfoldedClause> queryList, Clause currentQuery) {
		
		Set<Variable> bodyVariables = currentQuery.getVariables();
		bodyVariables.removeAll(currentQuery.getHead().getVariables());
		
		varLoop:
		for (Variable v : bodyVariables) {
			if (v.hasUnboundPrefix()) {
				continue;
			}
			
			int[] pattern = currentQuery.termIndexPattern(v);
			
			Set<Function> functionSet = null;
			Set<Variable> otherVariables = new HashSet<Variable>();
			Set<Constant> otherConstants = new HashSet<Constant>();

			for (int k = 0; k < pattern.length; k++) {
				if (pattern[k] == 0) {
					continue;
				}
				
				if (pattern[k] == 3) {
					continue varLoop;
				}
				
				Atom atom = currentQuery.getBodyAtomAt(k);
				AtomUnfoldList aur = au.getUnfolding(atom).unfoldList;
				
				Set<Function> newFunctions = null;
				
				if (atom.isConcept()) {
					newFunctions = aur.getFunctions();
				} else {
					Term otherTerm = null;
				
					if (pattern[k] == 1) {
						newFunctions = aur.getFunctions(0);
						if (!atom.getArgument(1).hasUnboundPrefix()) {
							otherTerm = atom.getArgument(1);
						}
					} else if (pattern[k] == 2) {
						newFunctions = aur.getFunctions(1);
						if (!atom.getArgument(0).hasUnboundPrefix()) {
							otherTerm = atom.getArgument(0);
						}
					}
					
					if (otherTerm != null) {
						if (otherTerm.isVariable()) {
							otherVariables.add((Variable)otherTerm);
						} else  if (otherTerm.isConstant()) {
							otherConstants.add((Constant)otherTerm);
						}
					}
				}
				
				if (otherConstants.size() > 1) {
					continue varLoop;
				}
				
				if (functionSet == null) {
					functionSet = new HashSet<Function>();
					functionSet.addAll(newFunctions);
				} else {
					functionSet = SetUtils.intersection(functionSet, newFunctions);
				}
				
				if (functionSet.isEmpty()) {
					continue varLoop;
				}
			}
			
			Set<Variable> allVariables = currentQuery.getVariables();
			
			if (functionSet != null) {

				Term substTerm = null;
				Substitution subst = new Substitution();
				
				if (otherConstants.size() == 0) {
					if (otherVariables.size() > 0) { 
						substTerm = otherVariables.iterator().next();
					} else {
						substTerm = Variable.getNewUnboundVariable(allVariables);
						allVariables.add((Variable)substTerm);
					}
				} else {
					substTerm = otherConstants.iterator().next();
				}

				for (Variable vv : otherVariables) {
					if (!vv.equals(substTerm)) {
						subst.add(vv, substTerm);
					}
				}

				ArrayList<Atom> baseBody = new ArrayList<Atom>();
				for (int k = 0; k < currentQuery.bodySize(); k++) {
					if (pattern[k] == 0) {
						baseBody.add(currentQuery.getBodyAtomAt(k).apply(subst));
					}
				}
		
				Set<Predicate> used = new HashSet<Predicate>();
				ArrayList<MarkedUnfoldedClause> newClauses = new ArrayList<MarkedUnfoldedClause>();

				++mark;
				
				for (Function f : functionSet) {

					Predicate newPred = lp.getPredicateForFunction(f);
					if (used.contains(newPred)) {
						continue;
					} else {
						used.add(newPred);
					}

					Atom newAtom;
					if (newPred.isConcept()) {
						newAtom = new Atom(newPred, substTerm);
					} else {
						Variable newVar = Variable.getNewUnboundVariable(allVariables);
//						allVariables.add(newVar);
						
						if (lp.getIndexForFunction(f) == 0) {
							newAtom = new Atom(newPred, substTerm, newVar);
						} else {
							newAtom = new Atom(newPred, newVar, substTerm);
						}
					}

					ArrayList<Atom> newBody = (ArrayList<Atom>)baseBody.clone();
					if (!newBody.contains(newAtom)) {
						newBody.add(newAtom);
					}

					Clause newClause = new Clause(currentQuery.getHead().apply(subst), newBody).condense();
					Substitution uSubst = newClause.prefixUnboundVariablesByUnderscoreSubstitution();
					
					UnfoldedClause newCC = new UnfoldedClause(newClause.apply(uSubst));
					
					if (newBody.size() > baseBody.size()) {
						newClauses.add(new MarkedUnfoldedClause(newCC, mark));
					} else {
						newClauses.clear();
						newClauses.add(new MarkedUnfoldedClause(newCC, mark));
						break;
					}
				}

				for (int i = 0; i < newClauses.size(); i++) {
					if (!queryList.contains(newClauses.get(i))) {
						queryList.add(newClauses.get(i));
					}
				}
				
			}
		}
	}

	private void unfold(ArrayList<MarkedUnfoldedClause> queryList) {
		
//		queryList = Clause.getNonRedundantClauses(queryList);

		int id = 0;
		
		int k = 0;
		while (k < queryList.size()) {
			
			MarkedUnfoldedClause muc = queryList.get(k++);
			Set<Atom> appendix = new HashSet<Atom>();
			int eIndex = -1;

			while (k < queryList.size() && queryList.get(k).mark == muc.mark) {
				Clause nc = queryList.get(k++).uClause.getClause();
				if (eIndex == -1) {
					eIndex = nc.bodySize() - 1;
				} else if (eIndex > -1 && nc.bodySize() - 1 != eIndex) {
					System.out.println("ERROR!!!");
				} 
				
				appendix.add(nc.getBodyAtomAt(eIndex));
			}
			
			ArrayList<Clause> last = new ArrayList<Clause>();
			
			Set<Variable> rootVars = muc.uClause.getClause().getBoundVariables();
			
			for (UnfoldedClause cp : au.unfold(muc.uClause, eIndex, appendix)) {
				Clause up = ((UnfoldedClause)cp.removeRedundantAtoms()).getClause();

				if (AtomUnfolder.mode == 1) {
					if (cp.isPure() && qc.restoreVariables(up).getVariables().containsAll(rootVars)) {
						qc.add(id, up);
					} else {
						last.add(up);
					}
				} else if (AtomUnfolder.mode == 2) {
					if (qc.restoreVariables(up).getVariables().containsAll(rootVars)) {
						qc.add(id, up);
					} else {
						last.add(up);
					}
				}
			}
			
			id++;
			for (Clause extraClause : last) {
				qc.add(id++, extraClause);
			}
		}
	}
	
	public void expand(ArrayList<MarkedUnfoldedClause> queryList) {
		ArrayList<Clause> queries = new ArrayList<Clause>();
		
		for (int k = 0; k < queryList.size(); k++) {
			queries.add(au.testUnfold(queryList.get(k).getClause()).prefixUnboundVariablesByUnderscore());
		}
		
//		queries = Clause.getNonRedundantClauses(queries);

		Variable v1 = new Variable("x");
		Variable v2 = new Variable("y");
		
		Map<Atom, Atom> atoms = new HashMap<Atom, Atom>();
		
		for (int k = 0; k < queries.size(); k++) {
			Clause c = queries.get(k);

			ArrayList<Atom> newBody = new ArrayList<Atom>();
			
			for (Atom atom : c.getBody()) {
				Predicate atomPred = atom.getPredicate();
				Predicate newPred = null;
				Atom clauseAtom = null;
				Atom unfoldAtom = null;
				Atom rootAtom = null;

				boolean u1 = false;
				boolean u2 = false;

				if (atom.getArgument(0).hasUnboundPrefix()) {
					u1 = true;
				}

				if (atom.isConcept()) {
					if (u1) {
						newPred = new Predicate(atomPred.getName() + "_u", 0);
						
						clauseAtom = new Atom(newPred, new Tuple());
						rootAtom = new Atom(newPred, new Tuple());
						unfoldAtom = new Atom(atomPred, new Tuple(Variable.UNBOUND_VARIABLE_1));
						
						atoms.put(rootAtom, unfoldAtom);
					} else {
//						newPred = atomPred;
						newPred = new Predicate(atomPred.getName() + "_.", 0);
						
						clauseAtom = new Atom(newPred, atom.getArguments());
						rootAtom = new Atom(newPred, new Tuple(v1));
						unfoldAtom = new Atom(atomPred, new Tuple(v1));
						
						atoms.put(rootAtom, unfoldAtom);
					}
				} else { 
					
					if (atom.getArguments().size() > 1 && atom.getArgument(1).hasUnboundPrefix()) {
						u2 = true;
					}
					
					if (u1 && u2) {
						newPred = new Predicate(atomPred.getName() + "_uu", 0);

						clauseAtom = new Atom(newPred, new Tuple());
						rootAtom = new Atom(newPred, new Tuple());
						unfoldAtom = new Atom(atomPred, new Tuple(Variable.UNBOUND_VARIABLE_1, Variable.UNBOUND_VARIABLE_2));
						
						atoms.put(rootAtom, unfoldAtom);

					} else if (u1) {
						newPred = new Predicate(atomPred.getName() + "_u.", 1);

						clauseAtom = new Atom(newPred, new Tuple(atom.getArgument(1)));
						rootAtom = new Atom(newPred, new Tuple(v2));
						unfoldAtom = new Atom(atomPred, new Tuple(Variable.UNBOUND_VARIABLE_1, v2));
						
						atoms.put(rootAtom, unfoldAtom);

					} else if (u2) {
						newPred = new Predicate(atomPred.getName() + "_.u", 1);

						clauseAtom = new Atom(newPred, new Tuple(atom.getArgument(0)));
						rootAtom = new Atom(newPred, new Tuple(v1));
						unfoldAtom = new Atom(atomPred, new Tuple(v1, Variable.UNBOUND_VARIABLE_2));
						
						atoms.put(rootAtom, unfoldAtom);
					} else {
//						newPred = atomPred;
						newPred = new Predicate(atomPred.getName() + "_..", 1);
						
						clauseAtom = new Atom(newPred, atom.getArguments());
						rootAtom = new Atom(newPred, new Tuple(v1, v2));
						unfoldAtom = new Atom(atomPred, new Tuple(v1, v2));
						
						atoms.put(rootAtom, unfoldAtom);
					}
				}
				
				newBody.add(clauseAtom);
			}

			qc.add(0, new Clause(query.getHead(), newBody));
		}
		
		
		Substitution s = new Substitution();
		s.add(Variable.UNBOUND_VARIABLE_1, Variable.UNBOUND_VARIABLE);
		s.add(Variable.UNBOUND_VARIABLE_2, Variable.UNBOUND_VARIABLE);
		
		for (Map.Entry<Atom, Atom> entry : atoms.entrySet()) {
			Atom a1 = entry.getKey();
			Atom a2 = entry.getValue();
			
			if (!a1.equals(a2)) {
				qc.add(0, new Clause(a1, a2).apply(s));
			}
			
			AtomUnfoldResult aur = au.getUnfolding(a2);
				
			Atom aa = a2.apply(aur.substitution);
			
			for (Atom b : aur.unfoldList.getAtoms()) {
				
				Atom bb = b.apply(aur.substitution);

				if (!aa.equals(bb)) {
					qc.add(0, new Clause(a1, bb).apply(s));
				}
			}
		}

	}

	private class MarkedUnfoldedClause implements ClauseContainer {
		private UnfoldedClause uClause;
		private int mark;
		
		public MarkedUnfoldedClause(UnfoldedClause uClause, int mark) {
			this.uClause = uClause;
			this.mark = mark;
		}
		
		public Clause getClause() {
			return uClause.getClause();
		}
		
		public String toString() {
			return mark + " " + uClause.getClause();
		}

		public <U extends ClauseContainer> boolean isCCSubsumedBy(U u) {
			return uClause.getClause().isSubsumedBy(u.getClause());
		}
		
		public int hashCode() {
			return uClause.hashCode();
		}
		
		public boolean equals(Object obj) {
			if (obj instanceof MarkedUnfoldedClause) {
				return uClause.equals(((MarkedUnfoldedClause)obj).uClause);
			} else {
				return false;
			}
		}
		
	}

}
