package edu.ntua.ivml.qa.db;

import java.text.NumberFormat;
import java.util.ArrayList;

import edu.ntua.ivml.qa.algorithm.Engine;
import edu.ntua.ivml.qa.dllite.OWL2ClauseSet;
import edu.ntua.ivml.qa.lp.Atom;
import edu.ntua.ivml.qa.lp.Clause;
import edu.ntua.ivml.qa.lp.Substitution;
import edu.ntua.ivml.qa.lp.Variable;
import edu.ntua.ivml.qa.utils.Utils;




public class QueryCollector {

	protected Clause initQuery;
	
	private ArrayList<Integer> ids;

	protected ArrayList<Clause> queries;
	protected ArrayList<Clause> nrQueries;
	
	protected ArrayList<Integer> nrBlockStart;
	protected ArrayList<Integer> nrBlockStartIds;
	
	protected boolean nrc;

	protected static NumberFormat nf = NumberFormat.getNumberInstance();
	
	public Substitution initSubst;
	
	static {
		nf.setMaximumFractionDigits(3);
		nf.setMinimumFractionDigits(3);
	}
	
	public QueryCollector(Clause initQuery) {
		this.initQuery = initQuery;
		
		initSubst = new Substitution();
		for (Variable v : initQuery.getVariables()) {
			initSubst.add(new Variable(Variable.UNBOUND_PREFIX + v.getName()), v);
		}
		
		nrBlockStart = new ArrayList<Integer>();
		nrBlockStartIds = new ArrayList<Integer>();
		
		ids = new ArrayList<Integer>();
		
		queries = new ArrayList<Clause>();
	}

	public Clause restoreVariables(Clause c) {
		return c.apply(initSubst);
	}
	
	public synchronized boolean add(Clause c) {
		return add(-1, c);
	}

	public synchronized boolean add(int id, Clause c) {
		for (Atom atom : c.getBody()) {
			if (atom.getPredicate().getName().startsWith(OWL2ClauseSet.AUX_PREDICATE)) {
				return false;
			}
		}
		
		if (Engine.removeRedundant && isRedundant(id, c)) {
			return false;
		}	

		Clause cc = c.apply(initSubst);
		
		queries.add(cc);

		if (id == -1) {
			ids = null;
		} else {
			if (nrBlockStart.size() == 0) {
				nrBlockStart.add(0);
				nrBlockStartIds.add(id);
			} else {
				if (nrBlockStartIds.get(nrBlockStart.size() - 1) != id) {
					nrBlockStart.add(ids.size());
					nrBlockStartIds.add(id);
				}
			}
			
			ids.add(id);
		}
		
		nrc = false;
		
		return true;
	}

	public ArrayList<Clause> getQueries() {
		return queries; 
	}
	
	public Substitution getInitSubstitution() {
		return initSubst;
	}
	
	public ArrayList<Clause> getNonRedundantQueries() {

		if (!nrc) {
			nrQueries = Clause.getNonRedundantClauses(queries, Engine.removeRedundant, ids, nrBlockStart, nrBlockStartIds);			
			nrc = true;
		} 
		
		return nrQueries;
	}

	private boolean isRedundant(int id, Clause query) {

		for (int i = 0; i < queries.size();) {
			if (ids != null && id == ids.get(i)) {
				i = nrBlockStart.get(Utils.binarySearch(nrBlockStartIds, id));
				if (i + 1 < nrBlockStart.size()) {
					i = nrBlockStart.get(i + 1);
				} else {
					break;
				}
				continue;
			}
			
			if (query.isSubsumedBy(queries.get(i))) {
				return true;
			}
			
			i++;
		}
		
		return false;
	}		
	
}
