package edu.ntua.ivml.qa.dllite;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import edu.ntua.ivml.qa.lp.Atom;
import edu.ntua.ivml.qa.lp.Clause;
import edu.ntua.ivml.qa.lp.Function;
import edu.ntua.ivml.qa.lp.FunctionalTerm;
import edu.ntua.ivml.qa.lp.LogicProgram;
import edu.ntua.ivml.qa.lp.Predicate;
import edu.ntua.ivml.qa.lp.Substitution;
import edu.ntua.ivml.qa.lp.Term;
import edu.ntua.ivml.qa.lp.Variable;


public class DLLiteRTheory extends LogicProgram {

	private Map<Function, Predicate> functionIndex;
	private Map<Function, Integer> posIndex;
	
	public DLLiteRTheory() {
		super();
		
		functionIndex = new HashMap<Function, Predicate>();
		posIndex = new HashMap<Function, Integer>();
	}
	
	public DLLiteRTheory(Collection<Clause> set) {		
		this();
		
		for (Clause c : set) {
			addClause(c);
		}
	}	
	
	public void addClause(Clause r) {

		Substitution subst = new Substitution();

		for (Variable v : r.getUnboundVariables()) {
			subst.add(v, Variable.UNBOUND_VARIABLE);
		}
		
		r = r.apply(subst);
		
		if (!rules.contains(r)) {
			rules.add(r);
			
			ArrayList<Clause> clauses = ruleIndex.get(r.getHead().getPredicate());
			if (clauses == null) {
				clauses = new ArrayList<Clause>();
				ruleIndex.put(r.getHead().getPredicate(), clauses);
			}
			clauses.add(r);
		}		

		Atom head = r.getHead();
		if (!head.isFunctionFree()) {
			
			Function f = null;
			
			if (head.isRole()) {
				for (int i = 0; i < 2; i++) {
					Term a = head.getArgument(i);
					
					if (a.isFunctionalTerm()) {
						f = ((FunctionalTerm)a).getFunction();
						if (functionIndex.containsKey(f)) {
							System.out.println("ERROR in DL-LiteR theory: Duplicate Function");
						} else {
							Atom body = r.getBodyAtomAt(0);
							functionIndex.put(f, body.getPredicate());
							
							if (i == 0) {
								posIndex.put(f, 1);
							} else {
								posIndex.put(f, 0);
							}
						}
					}
				}
			} else {
				f = ((FunctionalTerm)head.getArgument(0)).getFunction();
			}
		}
	}	

	public Predicate getPredicateForFunction(Function f) {
		return functionIndex.get(f);
	}
	
	public int getIndexForFunction(Function f) {
		return posIndex.get(f);
	}
}
