package edu.ntua.ivml.qa.dllite;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;


import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLAnnotationAssertionAxiom;
import org.semanticweb.owlapi.model.OWLAxiom;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassAssertionAxiom;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLDataFactory;
import org.semanticweb.owlapi.model.OWLDataPropertyAssertionAxiom;
import org.semanticweb.owlapi.model.OWLDeclarationAxiom;
import org.semanticweb.owlapi.model.OWLEquivalentClassesAxiom;
import org.semanticweb.owlapi.model.OWLEquivalentObjectPropertiesAxiom;
import org.semanticweb.owlapi.model.OWLInverseObjectPropertiesAxiom;
import org.semanticweb.owlapi.model.OWLObjectIntersectionOf;
import org.semanticweb.owlapi.model.OWLObjectInverseOf;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLObjectPropertyAssertionAxiom;
import org.semanticweb.owlapi.model.OWLObjectPropertyDomainAxiom;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectPropertyRangeAxiom;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLObjectUnionOf;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLPropertyExpression;
import org.semanticweb.owlapi.model.OWLSubClassOfAxiom;
import org.semanticweb.owlapi.model.OWLSubObjectPropertyOfAxiom;

import edu.ntua.ivml.qa.lp.Atom;
import edu.ntua.ivml.qa.lp.Clause;
import edu.ntua.ivml.qa.lp.Function;
import edu.ntua.ivml.qa.lp.FunctionalTerm;
import edu.ntua.ivml.qa.lp.Predicate;
import edu.ntua.ivml.qa.lp.Variable;


import uk.ac.manchester.cs.owl.owlapi.OWLDataFactoryImpl;

public class OWL2ClauseSet {

	public static String AUX_PREDICATE = "_AUX_";
	
	private int fcount;
	private int auxcount;

	private Set<Clause> rules;
	
	private ArrayList<OWLAxiom> ignoredAxioms;

	private OWLDataFactory factory;
	
	private OntologyReference ontRef;
	
	private boolean elhi = false;
	
	public OWL2ClauseSet(IRI uri) throws OWLOntologyCreationException {
		this(new OntologyReference(uri));	
	}

	public OWL2ClauseSet(OntologyReference ontRef) {
		this.ontRef = ontRef;	
	}
	
	public Set<Clause> rewrite() {
		fcount = 0;
		auxcount = 0;
		
		rules = new HashSet<Clause>();
		
		factory = new OWLDataFactoryImpl();

		ignoredAxioms = new ArrayList<OWLAxiom>();
		
		for (OWLAxiom axiom : ontRef.getOntology().getAxioms()) {
			processStr(axiom);
		}
		
		return rules;
	}

	
	private Atom createConceptAtom(OWLClass c, String v) {
		return new Atom(new Predicate(ontRef.handleIRI(c), 1), new Variable(v));
	}

	private Set<Atom> createFunctionalRoleAtoms(OWLObjectSomeValuesFrom ex, String v) {
		Set<Atom> atoms = new HashSet<Atom>();
		
		OWLObjectPropertyExpression p = ex.getProperty();
		
		if (p instanceof OWLObjectProperty) {
			atoms.add(new Atom(new Predicate(ontRef.handleIRI((OWLObjectProperty)p), 2), new Variable(v), new FunctionalTerm(new Function("f_" + fcount, 1), new Variable(v))));

		} else if (p instanceof OWLObjectInverseOf) {
			atoms.add(new Atom(new Predicate(ontRef.handleIRI((OWLObjectProperty)((OWLObjectInverseOf)p).getInverse()), 2), new FunctionalTerm(new Function("f_" + fcount, 1), new Variable(v)), new Variable(v)));
		}
	
		OWLClassExpression f = ex.getFiller();
		if (!f.isOWLThing() && f instanceof OWLClass) {
			atoms.add(new Atom(new Predicate(ontRef.handleIRI((OWLClass)f), 1), new FunctionalTerm(new Function("f_" + fcount, 1), new Variable(v))));
		}
		
		fcount++;
		
		return atoms;
	}
	
	private Atom createRoleAtom(OWLPropertyExpression<?, ?> p, String v1, String v2) {
		Atom atom = null;
		
		if (p instanceof OWLObjectProperty) {
			atom = new Atom(new Predicate(ontRef.handleIRI((OWLObjectProperty)p), 2), new Variable(v1), new Variable(v2));
		} else if (p instanceof OWLObjectInverseOf) {
			atom = new Atom(new Predicate(ontRef.handleIRI((OWLObjectProperty)((OWLObjectInverseOf)p).getInverse()), 2), new Variable(v2), new Variable(v1));
		}
		
		return atom;
	}
	
	public void add(Clause c) {
		rules.add(c);
	}
	
	public ArrayList<OWLAxiom> getIgnoredAxioms() {
		return ignoredAxioms;
	}
	
	private boolean processStr(OWLAxiom descr) {

//		if (!(descr instanceof OWLDeclarationAxiom)) {
//			System.out.println(descr.getClass()+ " " + descr);
//		}
		
		if (descr instanceof OWLEquivalentClassesAxiom) {
			boolean ret = false;
			for (OWLSubClassOfAxiom ax : ((OWLEquivalentClassesAxiom)descr).asOWLSubClassOfAxioms()) {
				if (processStr(ax)) {
					ret = true;
				} else {
					ignoredAxioms.add(descr);
				}
			}
			return ret;
			
		} else if (descr instanceof OWLSubClassOfAxiom) {
			OWLClassExpression subClass = ((OWLSubClassOfAxiom)descr).getSubClass();
			OWLClassExpression supClass = ((OWLSubClassOfAxiom)descr).getSuperClass();

			if (supClass.isOWLThing() || subClass.isOWLNothing()) {
				return true;
			} else if (supClass instanceof OWLObjectIntersectionOf) {
				boolean ret = false;
				for (OWLClassExpression ex : ((OWLObjectIntersectionOf)supClass).getOperands()) {
					if (processStr(factory.getOWLSubClassOfAxiom(subClass, ex))) {
						ret = true;
					} else {
						ignoredAxioms.add(descr);
					}
				}
				return ret;
			} 
			
			ArrayList<Atom> body = new ArrayList<Atom>();

			if (subClass instanceof OWLClass) {
				body.add(createConceptAtom((OWLClass)subClass, "x"));
			} else if (subClass instanceof OWLObjectSomeValuesFrom) {
				OWLObjectSomeValuesFrom cDescr = (OWLObjectSomeValuesFrom)subClass;
				OWLObjectPropertyExpression subProp = cDescr.getProperty();
				OWLClassExpression filler = cDescr.getFiller();
				
				if (!(filler.isOWLThing() || filler instanceof OWLClass)) {
					ignoredAxioms.add(descr);
					return false;
				} else {
					body.add(createRoleAtom(subProp, "x", "y"));
					if (!filler.isOWLThing()) {
						if (!elhi) {
							ignoredAxioms.add(descr);
							return false;
						}
						body.add(createConceptAtom((OWLClass)filler, "y"));
					}
				}
				
			} else if (subClass instanceof OWLObjectUnionOf) {
				boolean ret = false;
				for (OWLClassExpression ex : ((OWLObjectUnionOf)subClass).getOperands()) {
					if (processStr(factory.getOWLSubClassOfAxiom(ex, supClass))) {
						ret = true;
					} else {
						ignoredAxioms.add(descr);
					}
				}
				return ret;
				
			} else if (subClass instanceof OWLObjectIntersectionOf) {
				if (!elhi) {
					ignoredAxioms.add(descr);
					return false;
				}
				
				ArrayList<OWLClassExpression> classes = flattenClassIntersection((OWLObjectIntersectionOf)subClass);
				
				int j = 1;
				if (classes != null) {
					for (OWLClassExpression c : classes) {
						if (c instanceof OWLClass) {
							body.add(createConceptAtom((OWLClass)c, "x"));
						} else {
							OWLObjectSomeValuesFrom cDescr = (OWLObjectSomeValuesFrom)c;
							OWLObjectPropertyExpression subProp = cDescr.getProperty();
							OWLClassExpression filler = cDescr.getFiller();
							
							if (!(filler.isOWLThing() || filler instanceof OWLClass)) {
								ignoredAxioms.add(descr);
								return false;
							} else {
								body.add(createRoleAtom(subProp, "x", "y_" + j));
								if (!filler.isOWLThing()) {
									body.add(createConceptAtom((OWLClass)filler, "y_" + j));
								}
								j++;
							}
						}
					}
				} else {
					ignoredAxioms.add(descr);
					return false;
				}
			} else {
				ignoredAxioms.add(descr);
				return false;
			}
			
			Set<Atom> head = new HashSet<Atom>();
			
			if (supClass instanceof OWLClass) {
				head.add(createConceptAtom((OWLClass)supClass, "x"));
			} else if (supClass instanceof OWLObjectSomeValuesFrom) {
				OWLObjectSomeValuesFrom cDescr = (OWLObjectSomeValuesFrom)supClass;
				OWLObjectPropertyExpression subProp = cDescr.getProperty();
				OWLClassExpression filler = cDescr.getFiller();

				if (filler.isOWLThing() || filler instanceof OWLClass) {
					head.addAll(createFunctionalRoleAtoms((OWLObjectSomeValuesFrom)supClass, "x"));
				} else {
					OWLClass newClass = factory.getOWLClass(AUX_PREDICATE + auxcount++, ontRef.getPrefixFormat());
					
					if (processStr(factory.getOWLSubClassOfAxiom(newClass, filler))) {
						return processStr(factory.getOWLSubClassOfAxiom(subClass, factory.getOWLObjectSomeValuesFrom(subProp, newClass)));
					} else {
						ignoredAxioms.add(descr);
						return false;
					}
				}
				
			} else {
				ignoredAxioms.add(descr);
				return false;
			}
			
			for (Atom a: head) {
				add(new Clause(a, body));
			}
			return true;
			
		} else if (descr instanceof OWLEquivalentObjectPropertiesAxiom) {
			boolean ret = false;
			for (OWLSubObjectPropertyOfAxiom ax : ((OWLEquivalentObjectPropertiesAxiom)descr).asSubObjectPropertyOfAxioms()) {
//				processStr(ax);
				if (processStr(factory.getOWLSubObjectPropertyOfAxiom(ax.getSubProperty(), ax.getSuperProperty()))) {
					ret = true;
				} else {
					ignoredAxioms.add(descr);
				}
				
				if (processStr(factory.getOWLSubObjectPropertyOfAxiom(ax.getSuperProperty(), ax.getSubProperty()))) {
					ret = true;
				} else {
					ignoredAxioms.add(descr);
				}
			}
			return ret;
			
		} else if (descr instanceof OWLSubObjectPropertyOfAxiom) {
			
			OWLPropertyExpression<?, ?> subProp = ((OWLSubObjectPropertyOfAxiom)descr).getSubProperty();
			OWLPropertyExpression<?, ?> supProp = ((OWLSubObjectPropertyOfAxiom)descr).getSuperProperty();
		
			add(new Clause(createRoleAtom(supProp, "x", "y"), createRoleAtom(subProp, "x", "y")));				
			
			return true;
			
		} else if (descr instanceof OWLObjectPropertyDomainAxiom) {
			OWLPropertyExpression<?, ?> prop = ((OWLObjectPropertyDomainAxiom)descr).getProperty();
			OWLClassExpression domain = ((OWLObjectPropertyDomainAxiom)descr).getDomain();
			
			Atom body = createRoleAtom(prop, "x", "y");

			Set<Atom> head = new HashSet<Atom>();
			if (domain instanceof OWLClass) {
				head.add(createConceptAtom((OWLClass)domain, "x"));
			} else if (domain instanceof OWLObjectSomeValuesFrom) {
				if (!elhi) {
					OWLObjectSomeValuesFrom cDescr = (OWLObjectSomeValuesFrom)domain;
					OWLClassExpression filler = cDescr.getFiller();

					if (!filler.isOWLNothing()) {
						ignoredAxioms.add(descr);
						return false;
					}
				}
				
           		head.addAll(createFunctionalRoleAtoms((OWLObjectSomeValuesFrom)domain, "x"));
			} else if (domain instanceof OWLObjectIntersectionOf) {
				boolean ret = false;
				for (OWLClassExpression ex : ((OWLObjectIntersectionOf)domain).getOperands()) {
					if (processStr(factory.getOWLObjectPropertyDomainAxiom((OWLObjectPropertyExpression)prop, ex))) {
						ret = true;
					} else {
						ignoredAxioms.add(descr);
					}
				}
				return ret;
			} else {
				ignoredAxioms.add(descr);
				return false;
			}
			
			for (Atom a: head) {
				add(new Clause(a, body));
			}

			return true;
			
		} else if (descr instanceof OWLObjectPropertyRangeAxiom) {
			OWLPropertyExpression<?, ?> prop = ((OWLObjectPropertyRangeAxiom)descr).getProperty();
			OWLClassExpression range = ((OWLObjectPropertyRangeAxiom)descr).getRange();
			
			Atom body = createRoleAtom(prop, "y", "x");
					
			Set<Atom> head = new HashSet<Atom>();
			if (range instanceof OWLClass) {
				head.add(createConceptAtom((OWLClass)range, "x"));
			} else if (range instanceof OWLObjectSomeValuesFrom) {
				if (!elhi) {
					OWLObjectSomeValuesFrom cDescr = (OWLObjectSomeValuesFrom)range;
					OWLClassExpression filler = cDescr.getFiller();

					if (!filler.isOWLNothing()) {
						ignoredAxioms.add(descr);
						return false;
					}
				}

           		head.addAll(createFunctionalRoleAtoms((OWLObjectSomeValuesFrom)range, "x"));
			} else if (range instanceof OWLObjectIntersectionOf) {
				boolean ret = false;
				for (OWLClassExpression ex : ((OWLObjectIntersectionOf)range).getOperands()) {
					if (processStr(factory.getOWLObjectPropertyRangeAxiom((OWLObjectPropertyExpression)prop, ex))) {
						ret = true;
					} else {
						ignoredAxioms.add(descr);
					}
				}
				return ret;
				
			} else {
				ignoredAxioms.add(descr);
				return false;
			}

			for (Atom a: head) {
				add(new Clause(a, body));
			}
			
			return true;

		} else if (descr instanceof OWLInverseObjectPropertiesAxiom) {
			OWLObjectProperty first = ((OWLInverseObjectPropertiesAxiom)descr).getFirstProperty().asOWLObjectProperty();
			OWLObjectProperty second = ((OWLInverseObjectPropertiesAxiom)descr).getSecondProperty().asOWLObjectProperty();
			
			add(new Clause(createRoleAtom(first, "x", "y"), createRoleAtom(second, "y", "x")));
			add(new Clause(createRoleAtom(second, "x", "y"), createRoleAtom(first, "y", "x")));			
		
			return true;
			
		} else if (!(descr instanceof OWLClassAssertionAxiom) && !(descr instanceof OWLObjectPropertyAssertionAxiom) &&
			       !(descr instanceof OWLAnnotationAssertionAxiom) && !(descr instanceof OWLDataPropertyAssertionAxiom) &&
			       !(descr instanceof OWLDeclarationAxiom)) {
			ignoredAxioms.add(descr);
			return false;
		} else {
			return false;
		}

	}
	
	private ArrayList<OWLClassExpression> flattenClassIntersection(OWLObjectIntersectionOf inter) {
		ArrayList<OWLClassExpression> result = new ArrayList<OWLClassExpression>();
		
		flattenClassIntersection(inter, result);
		
		return result;
	}
	
	private void flattenClassIntersection(OWLObjectIntersectionOf inter, ArrayList<OWLClassExpression> result) {
		if (result == null) {
			return;
		}
		
		for (OWLClassExpression ex : inter.getOperands()) {
			if (ex instanceof OWLClass) {
				result.add(ex);
			} else if (ex instanceof OWLObjectSomeValuesFrom) {
				result.add(ex);
			} else if (ex instanceof OWLObjectIntersectionOf) {
				flattenClassIntersection((OWLObjectIntersectionOf)ex, result);
			} else {
				result = null;
				return;
			}
		}
	}
	
}
