package edu.ntua.ivml.qa.lp;

import java.util.ArrayList;

public class ClauseParser {

	private String IMPLIES_SYMBOL = "<-";
	private String VARIABLE_SYMBOL = Variable.VARIABLE_PREFIX;
	
	public ClauseParser() {}
	
	public Clause parseClause(String s) {
		
		int pos = s.indexOf(IMPLIES_SYMBOL);
		
		String headString = s.substring(0,pos).trim();
		String bodyString = s.substring(pos + 2).trim();
		
		Atom head = parseAtom(headString);
		
		ArrayList<String> v = split(bodyString);

		ArrayList<Atom> body = new ArrayList<Atom>();
		
		for (int i = 0; i < v.size(); i++) {
			Atom newAtom = parseAtom(v.get(i));
			if (!body.contains(newAtom)) {
				body.add(newAtom);
			}
		}

		return (new Clause(head, body)).condense();
		//return (new Clause(head, body));
	}
	
	public Atom parseAtom(String s) {

		int pos1 = s.indexOf("(");
		int pos2 = s.lastIndexOf(")");

		if (pos1 < 0) {
			return new Atom(new Predicate(s,0));
		} else {
			
			String pre = s.substring(0, pos1).trim();
			String arg = s.substring(pos1 + 1, pos2).trim();
						
			ArrayList<String> v = split(arg);
						
			ArrayList<Term> terms = new ArrayList<Term>();
			for (String c : v) {
				terms.add(parseTermString(c));
			}
			
			return new Atom(new Predicate(pre, terms.size()), (Term[])terms.toArray(new Term[] {}));
		}
	}	
	
	private Term parseTermString(String s) {
		
		int pos1 = s.indexOf("(");
		int pos2 = s.lastIndexOf(")");

		if (pos1 < 0) {
			if (s.startsWith(VARIABLE_SYMBOL)) {
				return new Variable(s.substring(1));
			} else {
				return new Constant(s);
			}
		} else {
			
			String fun = s.substring(0, pos1).trim();
			String arg = s.substring(pos1 + 1, pos2).trim();
			
			ArrayList<String> v = split(arg);
			
			ArrayList<Term> terms = new ArrayList<Term>();
			for (String c : v) {
				terms.add(parseTermString(c));
			}
			
			Function ff = new Function(fun, terms.size());
			
			return new FunctionalTerm(ff, (Term[])terms.toArray(new Term[] {}));
		}
	}
	
	
	private ArrayList<String> split(String s) {
		ArrayList<String> res = new ArrayList<String>();
		
		char[] arr = s.toCharArray();
	
		int c = 0;
		int start = 0;
		
		for (int i = 0; i < arr.length; i++) {
			boolean close = false;
	
			if (arr[i] == '(') {
				c++;
			} else if (arr[i] == ')') {
				c--;
				close = true;
			}
		
			if (c == 0 && arr[i] == ',') {
				res.add(s.substring(start, i).trim());
				i = i + 1;
				
				while (i < arr.length && (arr[i] == ' ')) {
					i++;
				}
				start = i;
				
			}
			if (c == 0 && close) {
				res.add(s.substring(start, i + 1).trim());
				i = i + 1;
				
				while (i < arr.length && (arr[i] == ' ' || arr[i] == ',')) {
					i++;
				}
				start = i;
			}
				
		}
		
		if (start <= arr.length - 1) {
			res.add(s.substring(start));
		}
		
		return res;
		
	}
	
}
