package edu.ntua.ivml.qa.lp;

import java.util.HashSet;
import java.util.Set;

public class Variable extends SimpleTerm {
	
	public static String VARIABLE_PREFIX = "?";
	public static String UNBOUND_PREFIX = "_";
	public static String UNBOUND_NEW_PREFIX = "_*";
	
	public static Variable UNBOUND_VARIABLE = new Variable("_");
	public static Variable UNBOUND_VARIABLE_1 = new Variable(Variable.UNBOUND_PREFIX + "1");
	public static Variable UNBOUND_VARIABLE_2 = new Variable(Variable.UNBOUND_PREFIX + "2");
	
	public Variable(String name) {
		super(name);
	}
	
	public void setName(String name) {
		this.name = name;
	}

	public Set<Variable> getVariables() {
		Set<Variable> res = new HashSet<Variable>();
		res.add(this);
		
		return res;
	}

	public boolean isConstant() {
		return false;
	}	
	
	public boolean isVariable() {
		return true;
	}
	
	public boolean isFunctionalTerm() {
		return false;
	}
	
	public boolean hasUnboundPrefix() {
		return name.startsWith(UNBOUND_PREFIX);
	}

	public boolean hasUnboundNewPrefix() {
		return name.startsWith(UNBOUND_NEW_PREFIX);
	}

	public Variable removeUnboudPrefix() {
		if (hasUnboundPrefix()) {
			return new Variable(name.substring(1));
		} else {
			return this;
		}
	}
	
	public boolean hasUnboundVars() {
		return hasUnboundPrefix();
	}
	
	public static Variable getNewDistinctVariable(Set<Variable> vars) {
		int i = 1;
		
		Variable v = new Variable("v_" + i);
		while (vars.contains(v)) {
			v = new Variable("v_" + (++i));
		}
		
		return v;
	}
	
	public static Variable getNewUnboundVariable(Set<Variable> vars) {
		int i = 1;
		
		Variable v2 = new Variable("_v_" + i);
		while (vars.contains(v2)) {
			i++;
			v2 = new Variable("_v_" + i);
		}
		
		return v2;
	}
	
    public int depth() {
    	return 0;
    }
    
	public Term apply(Substitution subst) {
		Term t = subst.getSubstitutionTerm(this);
		if (t != null) {
			return t;
		} else {
			return this;
		}
	}	
	
	public Object clone() {
		Variable v = (Variable)super.clone();
		v.name = name;
		return v;
	}		
	
	public int compareTo(Term t) {
		if (t instanceof Constant || t instanceof FunctionalTerm) {
			return -1;
		}

		return name.compareTo(((Variable)t).getName());
	}
	
	public boolean equals(Object obj) {
		if (!(obj instanceof Variable)) {
			return false;
		}
		
		return name.equals(((Variable)obj).getName());
	}
	
	public boolean unboundEquals(Term t) {
		if (t.isVariable() && this.hasUnboundPrefix() && ((Variable)t).hasUnboundPrefix()) {
			return true;
		}
		
		return equals(t);
	}
	
	public int hashCode() {
		return name.hashCode();
	}
	
	public String toString() {
		return VARIABLE_PREFIX + name;
	}	
	
	public String getStringPrefix() {
		return "";
	}
	
}
